/* caligula - cairo based gtk+ theme
 *
 * Copyright © 2005 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include "caligula-rand.h"
#include "caligula-draw-grassy.h"

static void
swoosh (cairo_t *cr, int x, int y, int width, int height)
{
    const double swoosh_alpha = 0.10;
    double top_x, top_y, right_x, right_y;

    top_x = x + caligula_rand_within (width / 2, width);
    top_y = y - caligula_rand_within (0, height * 2);
    right_x = x + caligula_rand_within (0, width);
    right_y = y + height + caligula_rand_within (0, height);

    cairo_new_path (cr);
    cairo_move_to (cr, top_x, top_y);
    cairo_curve_to (cr,
		    top_x + 5, top_y - 5,
		    right_x, right_y - 5,
		    right_x, right_y);

    cairo_curve_to (cr,
		    right_x, right_y - 5,
		    top_x - 5, top_y,
		    top_x, top_y);

    cairo_close_path (cr);
    cairo_set_source_rgba (cr, 0, 0, 0, swoosh_alpha);
    cairo_fill (cr);
}	

void
caligula_draw_grassy_box (cairo_t *cr, int x, int y, int width, int height)
{
    cairo_pattern_t *pattern;
    const double gray = 0.92;
    const double outline_gray = 0.50;

    cairo_save (cr);

    cairo_rectangle (cr, x, y, width, height);
#if 1
    cairo_clip (cr);
#endif

    cairo_set_source_rgb (cr, 1, 1, 1);
    cairo_fill (cr);

    /* Gradient fill */
    pattern = cairo_pattern_create_linear (x, y, x, y + height);
    cairo_pattern_add_color_stop_rgba (pattern, 0, 0, 0, 0, 1);
    cairo_pattern_add_color_stop_rgba (pattern, 0.035, gray, gray, gray, 1);
    cairo_pattern_add_color_stop_rgba (pattern, 0.25, 1, 1, 1, 1);
    cairo_pattern_add_color_stop_rgba (pattern, 0.45, gray, gray, gray, 1);
    cairo_pattern_add_color_stop_rgba (pattern, 0.91, 1, 1, 1, 1);
    cairo_pattern_add_color_stop_rgba (pattern, 1, 0, 0, 0, 0);
    cairo_set_source (cr, pattern);

    cairo_rectangle (cr, x, y, width, height);
    cairo_fill (cr);

    swoosh (cr, x, y, width, height);
    swoosh (cr, x, y, width, height);
    swoosh (cr, x, y, width, height);

    /* Outline */
    cairo_set_source_rgb (cr, outline_gray, outline_gray, outline_gray);
    cairo_set_line_width (cr, 1);
    cairo_rectangle (cr, x + 0.5, y + 0.5, width - 1, height - 1);
    cairo_stroke (cr);

    cairo_restore (cr);
}
