/* caligula - cairo based gtk+ theme
 *
 * Copyright © 2005 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

/*
  Draw things with soft shadows.

  By Carl Worth <cworth@redhat.com> following an original concept
  by Diana Fong <dfong@redhat.com>.
*/

#include "caligula-draw-shadowed.h"

void
caligula_draw_shadowed_inset_box (cairo_t *cr,
				  int x, int y, int width, int height,
				  double shadow_width, double shadow_height)
{
    cairo_pattern_t *shadow;

    cairo_translate (cr, x, y);

    /* The gradient I'm using here is a piece-wise linear
     * approximation I made by manually examining Diana's mockup and
     * progressively adding segments until the visual effect was
     * pleasing.
     *
     * Better would be to write a function to compute the shadow
     * falloff with some analytic expression.
     */

    /* Left side */
    shadow = cairo_pattern_create_linear (0, 0, shadow_width, 0);
    cairo_pattern_add_color_stop_rgba (shadow, 0.0, 0, 0, 0, 0.74);
    cairo_pattern_add_color_stop_rgba (shadow, 0.5, 0, 0, 0, 0.25);
    cairo_pattern_add_color_stop_rgba (shadow, 0.75, 0, 0, 0, 0.075);
    cairo_pattern_add_color_stop_rgba (shadow, 1.0, 0, 0, 0, 0.0);
    cairo_set_source (cr, shadow);
    cairo_rectangle (cr, 0, 0, shadow_width, height);
    cairo_fill (cr);
    cairo_pattern_destroy (shadow);

    /* The left and top shadows are drawn in two separate operations,
     * one OVER the other. This means we're not really getting the
     * right effect in the corner. Fixing this probably requires more
     * general "gradient" support in cairo so we can describe the
     * whole shadow in a single drawing operation. */
    
    /* Top */
    shadow = cairo_pattern_create_linear (0, 0, 0, shadow_height);
    cairo_pattern_add_color_stop_rgba (shadow, 0.0, 0, 0, 0, 0.74);
    cairo_pattern_add_color_stop_rgba (shadow, 0.5, 0, 0, 0, 0.25);
    cairo_pattern_add_color_stop_rgba (shadow, 0.75, 0, 0, 0, 0.075);
    cairo_pattern_add_color_stop_rgba (shadow, 1.0, 0, 0, 0, 0.0);
    cairo_set_source (cr, shadow);
    cairo_rectangle (cr, 0, 0, width, shadow_height);
    cairo_fill (cr);
    cairo_pattern_destroy (shadow);

    /* Put a one-pixel highlight on the right and bottom */
    cairo_set_source_rgb (cr, 1, 1, 1);
    cairo_rectangle (cr, width - 1, 0, 1, height);
    cairo_rectangle (cr, 0, height - 1, width, 1);
    cairo_fill (cr);
    
    /* Then a half-sized, half-as-dark shadow just inside the
     * highlight on the right and bottom.
     *
     * Diana's mockup has almost no decoration on the right and bottom
     * sides, but since GTK+ only gives us symmetric thickness values,
     * I think we need to at least draw something. What I'm doing here
     * is modeled something after the default GTK+ theme.
     */
    shadow_width /= 2.0;
    shadow_height /= 2.0;
    width -= 1;
    height -= 1;

    /* Right side */
    shadow = cairo_pattern_create_linear (width, 0, width - shadow_width, 0);
    cairo_pattern_add_color_stop_rgba (shadow, 0.0, 0, 0, 0, 0.37);
    cairo_pattern_add_color_stop_rgba (shadow, 0.5, 0, 0, 0, 0.125);
    cairo_pattern_add_color_stop_rgba (shadow, 0.75, 0, 0, 0, 0.0375);
    cairo_pattern_add_color_stop_rgba (shadow, 1.0, 0, 0, 0, 0.0);
    cairo_set_source (cr, shadow);
    cairo_rectangle (cr, width - shadow_width, 0, shadow_width, height);
    cairo_fill (cr);
    cairo_pattern_destroy (shadow);

    /* Bottom */
    shadow = cairo_pattern_create_linear (0, height, 0, height - shadow_height);
    cairo_pattern_add_color_stop_rgba (shadow, 0.0, 0, 0, 0, 0.37);
    cairo_pattern_add_color_stop_rgba (shadow, 0.5, 0, 0, 0, 0.125);
    cairo_pattern_add_color_stop_rgba (shadow, 0.75, 0, 0, 0, 0.0375);
    cairo_pattern_add_color_stop_rgba (shadow, 1.0, 0, 0, 0, 0.0);
    cairo_set_source (cr, shadow);
    cairo_rectangle (cr, 0, height - shadow_height, width, shadow_height);
    cairo_fill (cr);
    cairo_pattern_destroy (shadow);
}
