/* caligula - cairo based gtk+ theme
 *
 * Copyright © 2005 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include "caligula-rand.h"
#include "caligula-draw-swirly.h"

enum { LEFT, RIGHT };
#define STRENGTH 10

static void
draw_swoosh (cairo_t *cr, int side, int x, int y, int width, int height,
	     int top_x, int top_y, int left_x, int left_y,
	     int right_x, int right_y, int bottom_x, int bottom_y)
{
    cairo_pattern_t *pattern;

    cairo_save (cr);

    cairo_set_fill_rule (cr, CAIRO_FILL_RULE_EVEN_ODD);
    if (side == LEFT) {
	cairo_new_path (cr);
	cairo_move_to (cr, top_x, y);
	cairo_line_to (cr, right_x + 1, y);
	cairo_line_to (cr, right_x + 1, top_y - 1);
	cairo_line_to (cr, left_x - 1, top_y - 1);
	cairo_line_to (cr, left_x - 1, bottom_y + 1);
	cairo_line_to (cr, x, bottom_y + 1);
	cairo_line_to (cr, x, left_y);
	cairo_close_path (cr);
	cairo_clip (cr);
    }
    else {
	cairo_new_path (cr);
	cairo_move_to (cr, bottom_x, y + height);
	cairo_line_to (cr, x + width, right_y);
	cairo_line_to (cr, x + width, top_y - 1);
	cairo_line_to (cr, right_x + 1, top_y - 1);
	cairo_line_to (cr, right_x + 1, bottom_y + 1);
	cairo_line_to (cr, left_x - 1, bottom_y + 1);
	cairo_line_to (cr, left_x - 1, y + height);
	cairo_close_path (cr);
	cairo_clip (cr);
    }

    cairo_new_path (cr);
    cairo_move_to (cr, top_x, top_y);
    cairo_curve_to (cr,
		    top_x + STRENGTH, top_y,
		    right_x, right_y - STRENGTH,
		    right_x, right_y);

    cairo_curve_to (cr,
		    right_x, right_y + STRENGTH,
		    bottom_x + STRENGTH, bottom_y,
		    bottom_x, bottom_y);

    cairo_curve_to (cr,
		    bottom_x - STRENGTH, bottom_y,
		    left_x, left_y + STRENGTH,
		    left_x, left_y);

    cairo_curve_to (cr,
		    left_x, left_y - STRENGTH,
		    top_x - STRENGTH, top_y,
		    top_x, top_y);
    cairo_close_path (cr);

    top_y += 1;
    left_x += 4;
    right_x -= 4;
    bottom_y -= 1;

    cairo_move_to (cr, top_x, top_y);
    cairo_curve_to (cr,
		    top_x + STRENGTH, top_y,
		    right_x, right_y - STRENGTH,
		    right_x, right_y);

    cairo_curve_to (cr,
		    right_x, right_y + STRENGTH,
		    bottom_x + STRENGTH, bottom_y,
		    bottom_x, bottom_y);

    cairo_curve_to (cr,
		    bottom_x - STRENGTH, bottom_y,
		    left_x, left_y + STRENGTH,
		    left_x, left_y);

    cairo_curve_to (cr,
		    left_x, left_y - STRENGTH,
		    top_x - STRENGTH, top_y,
		    top_x, top_y);

    if (side == LEFT)
	pattern = cairo_pattern_create_linear (left_x, bottom_y, top_x, top_y);
    else
	pattern = cairo_pattern_create_linear (right_x, top_y,
					       bottom_x, bottom_y);

    cairo_pattern_add_color_stop_rgba (pattern, 0, 0, 0, 0.5, 0.5);
    cairo_pattern_add_color_stop_rgba (pattern, 1, 0, 0, 0.2, 1);
    cairo_set_source (cr, pattern);
    cairo_set_fill_rule (cr, CAIRO_FILL_RULE_EVEN_ODD);
    cairo_fill (cr);
    cairo_pattern_destroy (pattern);

    cairo_restore (cr);
}	

static void
swoosh (cairo_t *cr, int x, int y, int width, int height)
{
    int top_x, top_y, left_x, left_y, right_x, right_y, bottom_x, bottom_y;

    top_x = x + caligula_rand_within (15, 40);
    top_y = y - 2;
    left_x = x - 4;
    left_y = caligula_rand_within (y + height / 2 - 5, y + height / 2 + 5);
    right_x = left_x + caligula_rand_within (50, 60);
    right_y = caligula_rand_within (y + height / 2 - 5, y + height / 2 + 5);
    bottom_x = x + caligula_rand_within (10, 20);
    bottom_y = y + height - caligula_rand_within (0, 4);

    draw_swoosh (cr, LEFT, x, y, width, height,
		 top_x, top_y, left_x, left_y,
		 right_x, right_y, bottom_x, bottom_y);

    top_x = x + width - caligula_rand_within (10, 20);
    top_y = y + caligula_rand_within (0, 4);
    right_x = x + width + 4;
    right_y = y + caligula_rand_within (height / 2 - 5, height / 2 + 5);
    left_x = right_x - caligula_rand_within (50, 60);
    left_y = y + caligula_rand_within (height / 2 - 5, height / 2 + 5);
    bottom_x = x + width - caligula_rand_within (15, 40);
    bottom_y = y + height + 2;

    draw_swoosh (cr, RIGHT, x, y, width, height,
		 top_x, top_y, left_x, left_y,
		 right_x, right_y, bottom_x, bottom_y);
}	

void
caligula_draw_swirly_box (cairo_t *cr, int x, int y, int width, int height)
{
    cairo_pattern_t *pattern;
    const double gray = 0.92;
    const double outline_gray = 0.50;

    /* Gradient fill */
    pattern = cairo_pattern_create_linear (x, y, x, y + height);
    cairo_pattern_add_color_stop_rgba (pattern, 0, 0, 0, 0, 1);
    cairo_pattern_add_color_stop_rgba (pattern, 0.035, gray, gray, gray, 1);
    cairo_pattern_add_color_stop_rgba (pattern, 0.25, 1, 1, 1, 1);
    cairo_pattern_add_color_stop_rgba (pattern, 0.45, gray, gray, gray, 1);
    cairo_pattern_add_color_stop_rgba (pattern, 0.91, 1, 1, 1, 1);
    cairo_pattern_add_color_stop_rgba (pattern, 1, 0, 0, 0, 0);
    cairo_set_source (cr, pattern);

    cairo_rectangle (cr, x, y, width, height);
    cairo_fill (cr);

    /* Outline */
    cairo_set_source_rgb (cr, outline_gray, outline_gray, outline_gray);
    cairo_set_line_width (cr, 1);
    cairo_rectangle (cr, x + 0.5, y + 0.5, width - 1, height - 1);
    cairo_stroke (cr);

    swoosh (cr, x, y, width, height);
}
