/* CGE - cairo-gtk-engine

   Copyright © 2003 Julien Boulnois

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public
   License along with this program; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307 USA
*/

#include "cge.h"

#define DETAIL(xx) ((detail) && (!strcmp(xx, detail)))

/*#define DEBUG*/

static void
debug_msg (gchar *fun,
	   const gchar *detail,
	   gdouble x,
	   gdouble y, 
	   gdouble w,
	   gdouble h)
{
#ifdef DEBUG
    printf ("cge:%s(%s) at %ix%ix%ix%i\n",fun,detail,(int)x, (int)y, (int)w, (int)h);
#endif
}

static void
color_gdk_to_rgb (GdkColor color,gdouble *r,gdouble *g,gdouble *b);

static GdkColor
color_tint_black (GdkColor color,gdouble t);

/*
  static GdkColor
  color_tint_white (GdkColor color,gdouble t);
*/ 

static void
gdk_get_off (GdkDrawable *window,gint* x_off,gint* y_off);

static void
cge_gdk (cairo_t *cr, GdkDrawable *window);

static void
cge_draw_line (cairo_t *cr, GdkColor c, gdouble alpha, 
	       double x1, double y1, 
	       double x2, double y2, 
	       gboolean line_width);

static cge_style *
cge_get_style (GtkStyle *style,const gchar *detail);

static cairo_pattern_t *
cge_create_gradient (cairo_t *cr,
		     gint ncols,
		     GdkColor *colors,
		     gdouble alpha,
		     cge_direction d,
		     gint x_off,
		     gint y_off,
		     gint width,
		     gint height);

/*
  static void
  cge_draw_slider (cairo_t *cr,
  GdkColor color,
  gdouble alpha,
  gint x,gint y,gint w, gint h,
  gdouble line_width,
  gint line_count);
*/

static void
cge_draw_checked (cairo_t *cr,
		  GdkColor color,
		  gdouble alpha,
		  gint x,gint y,gint w, gint h,
		  gdouble line_width);

static void
cge_draw_arrow_from_rc (cairo_t *cr,
			GtkStyle * style,
			GtkStateType state_type,
			const gchar *detail,
			GdkColor c,
			gdouble alpha,
			gint x,gint y,gint w,gint h,
			gboolean fill,
			gdouble line_width,
			GtkArrowType at);

static void
cge_draw_rectangle (cairo_t *cr,
		    GdkColor c,
		    gdouble alpha,
		    gint x,
		    gint y,
		    gint w,
		    gint h);

static void
cge_draw_box (cairo_t *cr,
	      GdkColor c,
	      cairo_pattern_t *pattern,
	      gdouble alpha,
	      gdouble x,
	      gdouble y,
	      gdouble w,
	      gdouble h,
	      gdouble round_size,
	      gboolean fill,
	      gdouble line_width);

static void
cge_draw_circle (cairo_t *cr,
		 GdkColor c,
		 cairo_pattern_t *pattern,
		 gdouble alpha,
		 gint x,
		 gint y,
		 gdouble round_size,
		 gboolean fill,
		 gdouble line_width);

static void
cge_draw_box_gap (cairo_t *cr,
		  GdkColor c,
		  cairo_pattern_t *pattern,
		  gdouble alpha,
		  gdouble x,
		  gdouble y,
		  gdouble w,
		  gdouble h,
		  gdouble round_size,
		  gint gap_x,
		  gint gap_width,
		  gboolean fill,
		  gdouble line_width);

static void
cge_draw_notebook_tab (cairo_t *cr,
		       GdkColor c,
		       cairo_pattern_t *pattern,
		       gdouble alpha,
		       gdouble x,
		       gdouble y,
		       gdouble w,
		       gdouble h,
		       gdouble round_size,
		       gboolean fill,
		       gdouble line_width);

static cairo_pattern_t *
cge_create_gradient_from_rc (cairo_t *cr,
			     GtkStyle * style,
			     GtkStateType state_type,
			     const gchar *detail,
			     gint x_off,
			     gint y_off,
			     gint width,
			     gint height);

static void
cge_draw_filled_box_from_rc (cairo_t *cr,
			     GtkStyle * style,
			     GtkStateType state_type,
			     const gchar *detail,
			     gint x_off,
			     gint y_off,
			     gint x,
			     gint y,
			     gint width,
			     gint height);

static void
cge_draw_unfilled_box_from_rc (cairo_t *cr,
			       GtkStyle * style,
			       GtkStateType state_type,
			       const gchar *detail,
			       gint x,
			       gint y,
			       gint width,
			       gint height);

static void
cge_draw_line_from_rc (cairo_t *cr,
		       GtkStyle * style,
		       GtkStateType state_type,
		       const gchar *detail,
		       double x1,
		       double y1,
		       double x2,
		       double y2); 

static void
cge_style_init (CgeStyle * style);

static void
cge_style_class_init (CgeStyleClass * klass);

static void
draw_hline (GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    const gchar * detail, gint x1, gint x2, gint y);

static void
draw_vline (GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    const gchar * detail, gint y1, gint y2, gint x);

static void
draw_tab (GtkStyle *style,
	  GdkWindow *window,
	  GtkStateType state_type,
	  GtkShadowType shadow_type,
	  GdkRectangle *area,
	  GtkWidget *widget,
	  const gchar *detail,
	  gint x,
	  gint y,
	  gint width,
	  gint height);

static void
draw_shadow (GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     const gchar * detail,
	     gint x, gint y, gint width, gint height);

static void
draw_shadow_gap (GtkStyle               *style,
		 GdkWindow              *window,
		 GtkStateType            state_type,
		 GtkShadowType           shadow_type,
		 GdkRectangle           *area,
		 GtkWidget              *widget,
		 const gchar            *detail,
		 gint                    x,
		 gint                    y,
		 gint                    width,
		 gint                    height,
		 GtkPositionType         gap_side,
		 gint                    gap_x,
		 gint                    gap_width);

static void
draw_arrow (GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    const gchar * detail,
	    GtkArrowType arrow_type,
	    gint fill, gint x, gint y, gint width, gint height);

static void
draw_box (GtkStyle * style,
	  GdkWindow * window,
	  GtkStateType state_type,
	  GtkShadowType shadow_type,
	  GdkRectangle * area,
	  GtkWidget * widget,
	  const gchar * detail,
	  gint x, gint y, gint width, gint height);

static void
draw_box_gap (GtkStyle *style,
	      GdkWindow *window,
	      GtkStateType state_type,
	      GtkShadowType shadow_type,
	      GdkRectangle *area,
	      GtkWidget *widget,
	      const gchar *detail,
	      gint x,
	      gint y,
	      gint width,
	      gint height,
	      GtkPositionType gap_side,
	      gint gap_x,
	      gint gap_width);

#if 0
static void
draw_flat_box (GtkStyle * style,
	       GdkWindow * window,
	       GtkStateType state_type,
	       GtkShadowType shadow_type,
	       GdkRectangle * area,
	       GtkWidget * widget,
	       const gchar * detail,
	       gint x, gint y, gint width, gint height);
#endif

static void
draw_check (GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    const gchar * detail,
	    gint x, gint y, gint width, gint height);

static void
draw_option (GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     const gchar * detail,
	     gint x, gint y, gint width, gint height);

static void
draw_extension (GtkStyle * style,
		GdkWindow * window,
		GtkStateType state_type,
		GtkShadowType shadow_type,
		GdkRectangle * area,
		GtkWidget * widget,
		const gchar * detail,
		gint x,
		gint y,
		gint width,
		gint height, GtkPositionType gap_side);

static void
draw_slider (GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     const gchar * detail,
	     gint x,
	     gint y,
	     gint width, gint height, GtkOrientation orientation);

static void
draw_handle (GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     const gchar * detail,
	     gint x,
	     gint y,
	     gint width, gint height, GtkOrientation orientation);

static void
draw_resize_grip (GtkStyle * style, GdkWindow * window,
		  GtkStateType state_type, GdkRectangle * area,
		  GtkWidget * widget, const gchar * detail,
		  GdkWindowEdge edge, gint x, gint y,
		  gint width, gint height);

static GtkStyleClass *parent_class = NULL;

static cge_style *
cge_get_style (GtkStyle *style,const gchar *detail)
{
    gint i;

    for (i=1;i<MAX_STYLES;i++)       
    {
	cge_style *cstyle=&THEME_DATA (style)->styles[i];
	if (cstyle->detail!=NULL && detail!=NULL)     
	{
	    /* printf ("%s - %s\n",cstyle->detail,detail); */
	    if (g_ascii_strncasecmp (cstyle->detail, detail, strlen (detail))==0)  
	    {
		/* printf ("style %s found\n",detail); */
		return cstyle;
	    }
	}
    }

    return &THEME_DATA (style)->styles[0];
}


static void
color_gdk_to_rgb (GdkColor  color,gdouble *r,gdouble *g,gdouble *b)
{
    *r= (double)color.red/ (double)65535;
    *g= (double)color.green/ (double)65535;
    *b= (double)color.blue/ (double)65535;
}

static GdkColor color_tint_black (GdkColor color,gdouble t)
{
    GdkColor ncolor=color;
    ncolor.red=color.red*t;
    ncolor.green=color.green*t;
    ncolor.blue=color.blue*t;

    return ncolor;
}

/*
  static GdkColor color_tint_white (GdkColor color,gdouble t)
  {
  GdkColor ncolor=color;
  ncolor.red=color.red/t;
  ncolor.green=color.green/t;
  ncolor.blue=color.blue/t;

  return ncolor;
  }
*/

static void
gdk_get_off (GdkDrawable *window,gint* x_off,gint* y_off)
{
    GdkDrawable *real_drawable;
  
    if (GDK_IS_WINDOW (window))
    {  
	gdk_window_get_internal_paint_info (window, &real_drawable, x_off, y_off);  
    }
}

/* rename this to cge_init */
static void
cge_gdk (cairo_t *cr, GdkDrawable *window)
{
    Display *dpy;
    Drawable drawable;
    GdkDrawable *real_drawable;
    gint x_off=0, y_off=0;
  
    /* Unabstract X from GTK+ */
    if (GDK_IS_WINDOW (window))
    {  
	gdk_window_get_internal_paint_info (window, &real_drawable, &x_off, &y_off);  
	dpy = gdk_x11_drawable_get_xdisplay (real_drawable);
	drawable = gdk_x11_drawable_get_xid (real_drawable);
    }
    else
    {
	dpy = gdk_x11_drawable_get_xdisplay (window);
	drawable = gdk_x11_drawable_get_xid (window);
    }
    /* Ignore GTK+ and use Cairo for drawing. */
    cairo_set_target_drawable (cr, dpy, drawable);

    if (GDK_IS_WINDOW (window))
	cairo_translate (cr,-(double)x_off,-(double)y_off);
}

static void
cge_draw_line (cairo_t *cr, GdkColor c, gdouble alpha, 
	       double x1, double y1,
	       double x2, double y2, 
	       gboolean line_width)
{
    gdouble r,g,b;	
    cairo_save (cr);

    cairo_new_path (cr);
    cairo_set_line_width (cr, line_width);
    cairo_set_line_cap (cr,CAIRO_LINE_CAP_ROUND);
  
    cairo_move_to (cr, x1, y1);
    cairo_line_to (cr, x2, y2); 

    color_gdk_to_rgb (c, &r, &g, &b);
    cairo_set_source_rgba (cr, r, g, b, alpha);
  
    cairo_close_path (cr);
    cairo_stroke (cr);
    cairo_restore (cr);
}

static cairo_pattern_t *
cge_create_gradient (cairo_t *cr,
		     gint ncols,
		     GdkColor *colors,
		     gdouble alpha,
		     cge_direction d,
		     gint x_off,
		     gint y_off,
		     gint width,
		     gint height)
{
    cairo_pattern_t *pattern;
    gint i;

    if (d == CGE_DIRECTION_HORIZONTAL)
	pattern = cairo_pattern_create_linear (x_off, y_off,
					       x_off + width, y_off);
    else
	pattern = cairo_pattern_create_linear (x_off, y_off,
					       x_off, y_off + height);
					   
    for (i=0;i<ncols;i++)
    {
	gdouble r,g,b;	

	color_gdk_to_rgb (colors[i],&r,&g,&b);     

	cairo_pattern_add_color_stop (pattern, i * 1.0 / ncols, r, g, b, alpha);
    } 

    return pattern;
}

/*
  static void
  cge_draw_slider (cairo_t *cr,
  GdkColor color,
  gdouble alpha, 
  gint x,gint y,gint w, gint h,
  gdouble line_width,
  gint line_count)
  {
  gint i;

  cairo_save (cr);  
  
  cairo_translate (cr, (double)x, (double)y);
  for (i=0;i<line_count;i++)
  {
  cge_draw_line (cr,
  color, 
  alpha, 
  0, (h/line_count)*i,w, (h/line_count)*i, 
  line_width);
  }

  cairo_restore (cr);
  } 
*/

static void
cge_draw_checked (cairo_t *cr,
		  GdkColor color,
		  gdouble alpha, 
		  gint x,gint y,gint w, gint h,
		  gdouble line_width)
{
    cairo_save (cr);  
  
    cairo_translate (cr, (double)x, (double)y);

    cge_draw_line (cr,
		   color, 
		   alpha, 
		   0, 0,w, h, 
		   line_width);

    cge_draw_line (cr,
		   color, 
		   alpha, 
		   0, h,w, 0, 
		   line_width);
  
    cairo_restore (cr);  
} 

static void
cge_draw_arrow_from_rc (cairo_t *cr,
			GtkStyle * style,
			GtkStateType state_type,
			const gchar *detail,
			GdkColor c,
			gdouble alpha, 
			gint x,gint y,gint w,gint h,
			gboolean fill,
			gdouble line_width, 
			GtkArrowType at)
{
    gdouble r,g,b, rt;
    cairo_save (cr);
    cairo_new_path (cr);
    cairo_set_line_width (cr, line_width);
    cairo_set_line_join (cr, CAIRO_LINE_JOIN_ROUND);

    switch (THEME_DATA (style)->arrow_style)
    {      
    case CGE_ARROW_NORMAL:
    default:
	rt=h/4 - h/10;
	break;
    case CGE_ARROW_SIMPLE:
	rt=h/4;
	break;
    }

    switch (at)
    {
    case GTK_ARROW_UP:
	cairo_translate (cr, (double)x, (double)y);
	cairo_move_to (cr, (double)w/2, rt);
	break;
    case GTK_ARROW_DOWN:
	cairo_translate (cr, (double)x, (double)y);
	cairo_move_to (cr, (double)w/2, h-rt);
	cairo_rotate (cr, M_PI);          
	break;
    case GTK_ARROW_LEFT:
	cairo_translate (cr, (double)x, (double)y);
	cairo_move_to (cr, rt, (double)h/2);
	cairo_rotate (cr, -M_PI_2);    
	break;
    case GTK_ARROW_RIGHT:
	cairo_translate (cr, (double)x, (double)y);
	cairo_move_to (cr, (double)w-rt, (double)h/2);
	cairo_rotate (cr, M_PI_2);    
	break;

    }  

    switch (THEME_DATA (style)->arrow_style)
    {      
    case CGE_ARROW_NORMAL:
    default:
	cairo_rel_line_to (cr,  (double) w/2,   (double)h/2);
	cairo_rel_line_to (cr, - (double)w/2,  (double)h/5);
	cairo_rel_line_to (cr, - (double)w/2,  - (double)h/5);
	break;
    case CGE_ARROW_SIMPLE:
	cairo_rel_line_to (cr,  (double) w/2,   (double)h/2);      
	cairo_rel_line_to (cr, - (double)w,0);
	break;
    }
  
    /* printf ("%i\n",THEME_DATA (style)->arrow_style); */

    color_gdk_to_rgb (c, &r, &g, &b);
    cairo_set_source_rgba (cr, r, g, b, alpha);
    cairo_close_path (cr);    

    if (fill)
	cairo_fill (cr);
    else	
	cairo_stroke (cr);

    cairo_restore (cr);
}

static void
cge_draw_rectangle (cairo_t *cr,
		    GdkColor c,
		    gdouble alpha,
		    gint x,
		    gint y,
		    gint w, 
		    gint h)
{
    gdouble r,g,b;	

    cairo_save (cr);
    cairo_translate (cr, (double)x, (double)y);
    cairo_rectangle (cr,0,0, (double)w, (double)h);
    color_gdk_to_rgb (c, &r, &g, &b);  
    cairo_set_source_rgba (cr, r, g, b, alpha);
    cairo_fill (cr);
    cairo_restore (cr);
}

/* FIXME: need to be merged with cge_draw_box 
   adding a gboolean beveled parameter 
   whith that button can be beveled and rounded
*/
static void
cge_draw_beveled_box (cairo_t *cr,
		      GdkColor c1,
		      GdkColor c2,
		      gdouble alpha,
		      gint x,
		      gint y,
		      gint w, 
		      gint h, 
		      gdouble line_width)
{
    gdouble r,g,b;	

    cairo_save (cr);

    x += line_width;
    y += line_width;
    w -= line_width*2;
    h -= line_width*2; 
  
    cairo_save (cr);
    cairo_new_path (cr);
    cairo_set_line_width (cr, line_width/2);  
    cairo_translate (cr, (double)x, (double)y);
    cairo_move_to (cr, 0, (double)h);  
    cairo_rel_line_to (cr,  0,   - (double)h);
    cairo_rel_line_to (cr,  (double)w,   0);
    color_gdk_to_rgb (c2,&r,&g,&b);
    cairo_set_source_rgba (cr, r, g, b, alpha);
    cairo_stroke (cr);  
    cairo_restore (cr);

    cairo_save (cr);
    cairo_new_path (cr);
    cairo_set_line_width (cr, line_width);  
    cairo_translate (cr, (double)x, (double)y);
    cairo_move_to (cr, line_width/2, (double)h+line_width/2);  
    cairo_rel_line_to (cr,  0,   - (double)h);
    cairo_rel_line_to (cr,  (double)w,   0);
    color_gdk_to_rgb (c1,&r,&g,&b);
    cairo_set_source_rgba (cr, r, g, b, alpha);
    cairo_stroke (cr);  
    cairo_restore (cr);

    cairo_save (cr);
    cairo_new_path (cr);
    cairo_set_line_width (cr, line_width);  
    cairo_translate (cr, (double)x, (double)y);
    cairo_move_to (cr, 0, (double)h);  
    cairo_rel_line_to (cr,  (double)w,   0);
    cairo_rel_line_to (cr,  0,   - (double)h);
    color_gdk_to_rgb (c2,&r,&g,&b);
    cairo_set_source_rgba (cr, r, g, b, alpha);
    cairo_stroke (cr);  
    cairo_restore (cr);
    /*
      cairo_save (cr);
      cairo_new_path (cr);
      cairo_set_line_width (cr, line_width);  
      cairo_translate (cr, (double)x, (double)y);
      cairo_move_to (cr, line_width/2, (double)h+line_width/2);  
      cairo_rel_line_to (cr,  (double)w,   0);
      cairo_rel_line_to (cr,  0,   - (double)h);
      color_gdk_to_rgb (c2,&r,&g,&b);
      cairo_set_source_rgba (cr, 0, 0, 0, alpha);
      cairo_stroke (cr);  
      cairo_restore (cr);
    */
    cairo_restore (cr);
}

static void
cge_draw_box (cairo_t *cr,
	      GdkColor c,
	      cairo_pattern_t *pattern,
	      gdouble alpha,
	      gdouble x,
	      gdouble y,
	      gdouble w, 
	      gdouble h, 
	      gdouble round_size, 
	      gboolean fill,
	      gdouble line_width)
{
    gdouble r,g,b;	

    cairo_save (cr);
       
    cairo_new_path (cr);
    cairo_set_line_width (cr, line_width);
    cairo_set_line_join (cr, CAIRO_LINE_JOIN_ROUND);

    if (!fill){
	x += line_width/2;
	y += line_width/2;
	w -= line_width;
	h -= line_width; 
    }

    cairo_translate (cr, (double)x, (double)y);
    cairo_move_to (cr, 0, (double)round_size);
    cairo_rel_curve_to (cr,
			0, - (double)round_size,
			(double)round_size/2, - (double)round_size,
			(double)round_size, - (double)round_size);

    cairo_rel_line_to (cr,  (double) w- (double)2*round_size,   0);

    cairo_rel_curve_to (cr,
			(double)round_size,0,
			(double)round_size, (double)round_size/2,
			(double)round_size, (double)round_size);

    cairo_rel_line_to (cr,   0,  (double)h- (double)2*round_size);
    cairo_rel_curve_to (cr,
			0, (double)round_size,
			- (double)round_size/2, (double)round_size,
			- (double)round_size, (double)round_size);

    cairo_rel_line_to (cr, - (double)w+ (double)2*round_size,   0); 
    cairo_rel_curve_to (cr,
			- (double)round_size,0,
			- (double)round_size,- (double)round_size/2,
			- (double)round_size,- (double)round_size);

    cairo_rel_line_to (cr, 0,   - (double)h+ (double)2*round_size); 

    if (pattern != NULL) 
	cairo_set_source (cr, pattern);
    else {
	color_gdk_to_rgb (c, &r, &g, &b);
	cairo_set_source_rgba (cr, r, g, b, alpha);
    }

    /* cairo_close_path (cr); */

    if (fill)
	cairo_fill (cr);
    else	
	cairo_stroke (cr);

    cairo_restore (cr);
}

static void
cge_draw_circle (cairo_t *cr,
		 GdkColor c,
		 cairo_pattern_t *pattern,
		 gdouble alpha,
		 gint x,
		 gint y,
		 gdouble round_size, 
		 gboolean fill,
		 gdouble line_width)
{
    gdouble r,g,b;	

    cairo_save (cr);

    if (!fill)
	round_size -=line_width/2;
  
    cairo_new_path (cr);
    cairo_set_line_width (cr, line_width);
    cairo_set_line_join (cr, CAIRO_LINE_JOIN_ROUND);
    /* cairo_set_tolerance (cr, 5); */

    cairo_translate (cr, (double)x, (double)y);
    cairo_move_to (cr, 0, (double)round_size);
    cairo_rel_curve_to (cr,
			0, - (double)round_size,
			(double)round_size/2, - (double)round_size,
			(double)round_size, - (double)round_size);

    cairo_rel_curve_to (cr,
			(double)round_size,0,
			(double)round_size, (double)round_size/2,
			(double)round_size, (double)round_size);

    cairo_rel_curve_to (cr,
			0, (double)round_size,
			- (double)round_size/2, (double)round_size,
			- (double)round_size, (double)round_size);

    cairo_rel_curve_to (cr,
			- (double)round_size,0,
			- (double)round_size,- (double)round_size/2,
			- (double)round_size,- (double)round_size);

    if (pattern != NULL) 
	cairo_set_source (cr, pattern);
    else {
	color_gdk_to_rgb (c, &r, &g, &b);
	cairo_set_source_rgba (cr, r, g, b, alpha);
    }
  
    cairo_close_path (cr);

    if (fill)
	cairo_fill (cr);
    else	
	cairo_stroke (cr);

    cairo_restore (cr);
}

/* FIXME : work only with top gap*/
static void
cge_draw_box_gap (cairo_t *cr,
		  GdkColor c,
		  cairo_pattern_t *pattern,
		  gdouble alpha,
		  gdouble x,
		  gdouble y,
		  gdouble w, 
		  gdouble h, 
		  gdouble round_size, 
		  gint gap_x,
		  gint gap_width,
		  gboolean fill,
		  gdouble line_width)
{
    gdouble r,g,b;	

    cairo_save (cr);
    cairo_new_path (cr);
    cairo_set_line_width (cr, line_width);
    cairo_set_line_join (cr, CAIRO_LINE_JOIN_ROUND);
    cairo_set_line_cap (cr,CAIRO_LINE_CAP_SQUARE);

    if (!fill) {
	x += line_width/2;
	y += line_width/2;
	w -= line_width;
	h -= line_width;
    } 

    cairo_translate (cr, (double)x, (double)y);
    cairo_move_to (cr, 0,0);
    cairo_rel_line_to (cr,  (double)gap_x,   0);

    cairo_move_to (cr, (double)gap_x+ (double)gap_width-line_width,0);
    cairo_rel_line_to (cr,  (double) w- (double)round_size- (double)gap_width- (double)gap_x+line_width,   0);

    cairo_rel_curve_to (cr,
			(double)round_size,0,
			(double)round_size, (double)round_size/2,
			(double)round_size, (double)round_size);

    cairo_rel_line_to (cr,   0,  (double)h- (double)2*round_size);
    cairo_rel_curve_to (cr,
			0, (double)round_size,
			- (double)round_size/2, (double)round_size,
			- (double)round_size, (double)round_size);

    cairo_rel_line_to (cr, - (double)w+ (double)2*round_size,   0); 
    cairo_rel_curve_to (cr,
			- (double)round_size,0,
			- (double)round_size,- (double)round_size/2,
			- (double)round_size,- (double)round_size);

    cairo_rel_line_to (cr, 0,   - (double)h); 

    if (pattern != NULL) 
	cairo_set_source (cr, pattern);
    else {
	color_gdk_to_rgb (c, &r, &g, &b);
	cairo_set_source_rgba (cr, r, g, b, alpha);
    }
  
    if (fill)
	cairo_fill (cr);
    else	
	cairo_stroke (cr);
    cairo_restore (cr);
}

/* FIXME : work only with top notebook tab*/
static void
cge_draw_notebook_tab (cairo_t *cr,
		       GdkColor c,
		       cairo_pattern_t *pattern,
		       gdouble alpha,
		       gdouble x,
		       gdouble y,
		       gdouble w, 
		       gdouble h, 
		       gdouble round_size, 
		       gboolean fill,
		       gdouble line_width)
{
    gdouble r,g,b;	

    cairo_save (cr);
    cairo_new_path (cr);
    cairo_set_line_width (cr, line_width);
    cairo_set_line_join (cr, CAIRO_LINE_JOIN_ROUND);
    cairo_set_line_cap (cr,CAIRO_LINE_CAP_SQUARE);

    if (!fill) 
    {
	x += line_width/2;
	y += line_width/2;
	w -= line_width;
	h -= line_width;
    } 

    cairo_translate (cr, (double)x, (double)y);
    cairo_move_to (cr, 0, (double)h);
    cairo_rel_line_to (cr,  0,  - (double)h+ (double)round_size );
    cairo_rel_curve_to (cr,
			0, - (double)round_size,
			(double)round_size/2, - (double)round_size,
			(double)round_size, - (double)round_size);

    cairo_rel_line_to (cr,  (double) w- (double)2*round_size,   0);
    cairo_rel_curve_to (cr,
			(double)round_size,0,
			(double)round_size, (double)round_size/2,
			(double)round_size, (double)round_size);

    cairo_rel_line_to (cr,   0,  (double)h- (double)round_size);

    if (pattern != NULL)
	cairo_set_source (cr, pattern);
    else {
	color_gdk_to_rgb (c, &r, &g, &b);
	cairo_set_source_rgba (cr, r, g, b, alpha);
    }

    if (fill)
	cairo_fill (cr);
    else	
	cairo_stroke (cr);
    cairo_restore (cr);
}

cairo_pattern_t *
cge_create_gradient_from_rc (cairo_t *cr,
			     GtkStyle * style,
			     GtkStateType state_type,
			     const gchar *detail,
			     gint x_off,
			     gint y_off,
			     gint width,
			     gint height)
{
    cairo_pattern_t *gradient;
    GdkColor *colors;
    gint i;
    cge_style *cstyle;
    cstyle=cge_get_style (style,detail);

    /* printf ("cge-draw:%i\n",cstyle->fill_style.gradient_colors_count[state_type]); */

    colors = g_new0 (GdkColor, cstyle->fill_style.gradient_colors_count[state_type]);

    if (cstyle->fill_style.fill_mode==CGE_FILL_SOLID) {
	g_free (colors);
	return NULL;
    }
  
    if (cstyle->fill_style.gradient_colors_default[state_type]==TRUE)
    {
	switch (state_type)
	{
	case GTK_STATE_PRELIGHT:
	    colors[0]=style->bg[GTK_STATE_PRELIGHT];
	    colors[1]=style->bg[GTK_STATE_SELECTED];
	  
	    break;
	default:	  
	    colors[0]=style->bg[state_type];
	    colors[1]=color_tint_black (style->bg[state_type],0.96);      
	}
    }
    else
    {
	for (i=0;i<cstyle->fill_style.gradient_colors_count[state_type];i++)
	{
	    GdkColor tmp=colors[i]=cstyle->fill_style.gradient_colors[state_type][i];
	  
	    colors[i]=tmp;
	}
    }
  
    if (cstyle->fill_style.fill_mode==CGE_FILL_GRADIENT)
    {

	gradient=cge_create_gradient (cr,
				      cstyle->fill_style.gradient_colors_count[state_type],
				      colors,
				      cstyle->fill_style.alpha,
				      cstyle->fill_style.gradient_direction,
				      x_off,y_off,width,height);
	g_free (colors);
	return gradient;
    }

    return NULL;
}


static void
cge_draw_filled_box_from_rc (cairo_t *cr,
			     GtkStyle * style,
			     GtkStateType state_type,
			     const gchar *detail,
			     gint x_off,
			     gint y_off,
			     gint x,
			     gint y,
			     gint width,
			     gint height)

{
    cairo_pattern_t *gradient;
    GdkColor color;
    cge_style *cstyle;

    cstyle=cge_get_style (style,detail);

    gradient=cge_create_gradient_from_rc (cr, style,state_type,detail,x_off,y_off,width,height); 

    if (DETAIL ("entry_bg") || DETAIL ("cell_odd") || DETAIL ("cell_even")) {
	color=style->base[state_type];
    
	cge_draw_rectangle (cr,color,1,x,y,width,height);
    }
    else if (DETAIL ("background"))
    {
	color=style->bg[state_type];
	cge_draw_box (cr,
		      color,
		      /* NULL, */
		      gradient,
		      1,
		      x,
		      y,
		      width,
		      height,
		      0,
		      TRUE,0);
    }
    else
    {
	if (cstyle->line_style.line_mode==CGE_LINE_BEVEL)
	{
	    /*	  
		 cairo_save (cr);
		 cairo_translate (cr, (double)x, (double)y);
		 cairo_rectangle (cr,0,0, (double)width, (double)height);
		 if (gradient!=NULL)
		 cairo_set_source (cr,gradient);
		 cairo_fill (cr);
		 cairo_restore (cr);
	    */	  
	    color=style->bg[state_type];
	  
	    cge_draw_box (cr,
			  color,
			  NULL,
			  cstyle->fill_style.alpha,
			  x,
			  y,
			  width,
			  height,
			  0,
			  TRUE,0);
	}
	else	
	{
	    color=style->bg[state_type];

	    cge_draw_box (cr,
			  color,
			  gradient,
			  cstyle->fill_style.alpha,
			  x,
			  y,
			  width,
			  height,
			  cstyle->fill_style.round_width,
			  TRUE,0);
	}    
    }

    if (gradient)
	cairo_pattern_destroy (gradient);
}



static void
cge_draw_unfilled_box_from_rc (cairo_t *cr,
			       GtkStyle * style,
			       GtkStateType state_type,
			       const gchar *detail,
			       gint x,
			       gint y,
			       gint width,
			       gint height)

{
    GdkColor color;
    cge_style *cstyle;

    cairo_save (cr);

    cstyle=cge_get_style (style,detail);

    if (cstyle->line_style.color_default[state_type]==TRUE)
	color=style->dark[state_type];
    else
	color=cstyle->line_style.color[state_type];

    if (cstyle->line_style.line_mode==CGE_LINE_DASHED)
    {
	/* for now, static dash */
	double d[2]={4,4};
	cairo_set_dash (cr, d, 2, 0);
    }
  
    if (cstyle->line_style.line_mode==CGE_LINE_BEVEL)
    {
	GdkColor light_color;
	GdkColor dark_color;
	switch (state_type)
	{
	case GTK_STATE_NORMAL:
	    light_color=style->light[state_type];
	    dark_color=style->dark[state_type];
	    break;
	case GTK_STATE_PRELIGHT:
	    light_color=style->bg[state_type];
	    dark_color=style->dark[state_type];
	    break;
	case GTK_STATE_ACTIVE:
	    light_color=style->dark[state_type];
	    dark_color=style->light[state_type];
	    break;
	      
	default:	  
	    light_color=style->light[state_type];
	    dark_color=style->dark[state_type];
	    break;
	}
      
	if (DETAIL ("scrolled_window"))
	{
	    light_color=style->dark[state_type];
	    dark_color=style->light[state_type];
	}
	cge_draw_beveled_box (cr,			  
			      light_color,
			      dark_color,
			      cstyle->line_style.alpha,
			      x,
			      y,
			      width,
			      height,
			      cstyle->line_style.line_width);
    }
    else if (cstyle->line_style.line_mode==CGE_LINE_SOLID)
    {
	/* Cannot have rounded box for menu */
	if (DETAIL ("menu"))
	    cge_draw_box (cr,
			  color,
			  NULL,
			  cstyle->line_style.alpha,
			  x,
			  y,
			  width,
			  height,
			  0,
			  FALSE,
			  cstyle->line_style.line_width);
	else
	    cge_draw_box (cr,
			  color,
			  NULL,
			  cstyle->line_style.alpha,
			  x,
			  y,
			  width,
			  height,
			  cstyle->line_style.round_width,
			  FALSE,
			  cstyle->line_style.line_width);
      
    }
    cairo_restore (cr);
}

static void
cge_draw_line_from_rc (cairo_t *cr,
		       GtkStyle * style,
		       GtkStateType state_type,
		       const gchar *detail,
		       gdouble x1,
		       gdouble y1,
		       gdouble x2,
		       gdouble y2)

{
    GdkColor color;
    cge_style *cstyle;

    cstyle=cge_get_style (style,detail);

    if (cstyle->line_style.color_default[state_type]==TRUE)
	color=style->dark[state_type];
    else
	color=cstyle->line_style.color[state_type];

    cairo_save (cr);
      
    if (cstyle->line_style.line_mode==CGE_LINE_DASHED)
    {
	/* for now, static dash */
	double d[2]={4,4};
	cairo_set_dash (cr, d, 2, 0);
    }

    cge_draw_line (cr,
		   color,
		   1,
		   x1,
		   y1,
		   x2,
		   y2,
		   cstyle->line_style.line_width);

    cairo_restore (cr);
}

static void
draw_hline (GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    const gchar * detail, gint x1, gint x2, gint y)
{
    cairo_t *cr;

    debug_msg ("draw_hline",detail,x1,y,x2,y);

    cr=cairo_create ();			
    cge_gdk (cr,window);

    cge_draw_line_from_rc (cr,
			   style,
			   state_type,
			   detail,
			   x1,
			   y+0.5,
			   x2,
			   y+0.5);

    cairo_destroy (cr);
}

static void
draw_vline (GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    const gchar * detail, gint y1, gint y2, gint x)
{
    cairo_t *cr;

    debug_msg ("draw_vline",detail,x,y1,x,y2);

    cr=cairo_create ();			
    cge_gdk (cr,window);

    cge_draw_line_from_rc (cr,
			   style,
			   state_type,
			   detail,
			   x+0.5,
			   y1,
			   x+0.5,
			   y2);

    cairo_destroy (cr);
}

static void
draw_tab (GtkStyle *style,
	  GdkWindow *window,
	  GtkStateType state_type,
	  GtkShadowType shadow_type,
	  GdkRectangle *area,
	  GtkWidget *widget,
	  const gchar *detail,
	  gint x,
	  gint y,
	  gint width,
	  gint height)
{
    cairo_t *cr;

    debug_msg ("draw_tab",detail,x,y,width,height);

    cr=cairo_create ();			
    cge_gdk (cr,window);


    cge_draw_arrow_from_rc (cr,
			    style,
			    state_type,
			    detail,
			    style->text[state_type],
			    1,
			    x+1,y,width-2, (height-2)/2,
			    TRUE,1.5,GTK_ARROW_UP);
    cge_draw_arrow_from_rc (cr,
			    style,
			    state_type,
			    detail,
			    style->text[state_type],
			    1,
			    x+1,y+3+ (height-2)/2,width-2, (height-2)/2,
			    TRUE,1.5,GTK_ARROW_DOWN);

    cairo_destroy (cr);
}

static void
draw_shadow (GtkStyle * style, 
	     GdkWindow * window,
	     GtkStateType state_type, 
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget, 
	     const gchar * detail,
	     gint x, 
	     gint y, 
	     gint width, 
	     gint height)

{
    cairo_t *cr;
    cge_style *cstyle;

    debug_msg ("draw_shadow",detail,x,y,width,height);

    cstyle=cge_get_style (style,detail);

    cr=cairo_create ();			

    cge_gdk (cr,window);
  
    if (DETAIL ("entry")) {
	cge_draw_box (cr,
		      style->dark[state_type],
		      NULL,
		      1,
		      x,y,width,height,
		      0,
		      FALSE,
		      cstyle->line_style.line_width);
	/* cge_draw_rectangle (cr,style->bg[state_type],1,x,y,width,height); */
    } else {
  	cge_draw_unfilled_box_from_rc (cr,
				       style,
				       state_type,
				       detail,
				       x,
				       y,
				       width,
				       height);
    }
  
    cairo_destroy (cr);
}

static void
draw_shadow_gap (GtkStyle               *style,
		 GdkWindow              *window,
		 GtkStateType            state_type,
		 GtkShadowType           shadow_type,
		 GdkRectangle           *area,
		 GtkWidget              *widget,
		 const gchar            *detail,
		 gint                    x,
		 gint                    y,
		 gint                    width,
		 gint                    height,
		 GtkPositionType         gap_side,
		 gint                    gap_x,
		 gint                    gap_width)
{
    cairo_t *cr;
    gint x_off=0,y_off=0;
    GdkColor color;
    cairo_pattern_t *gradient;
    cge_style *cstyle;

    debug_msg ("draw_shadow_gap",detail,x,y,width,height);

    cstyle=cge_get_style (style,detail);

    cr=cairo_create ();			

    cge_gdk (cr,window);

    gdk_get_off (window,&x_off,&y_off);

    gradient=cge_create_gradient_from_rc (cr, 
					  style,
					  state_type,
					  detail,
					  x_off,y_off,width,height); 
  
    if (cstyle->line_style.color_default[state_type]==TRUE)
	color=style->dark[state_type];
    else
	color=cstyle->line_style.color[state_type];

    cge_draw_box_gap (cr,
		      style->bg[state_type],
		      NULL,
		      cstyle->fill_style.alpha,
		      x,y,width,height,
		      cstyle->fill_style.round_width,
		      gap_x,gap_width,
		      TRUE,0);

    cge_draw_box_gap (cr,
		      color,
		      NULL,
		      cstyle->line_style.alpha,
		      x,y,width,height,
		      cstyle->line_style.round_width,
		      gap_x,gap_width,
		      FALSE,
		      cstyle->line_style.line_width);

    if (gradient)
	cairo_pattern_destroy (gradient);
  
    cairo_destroy (cr);
}

static void
draw_arrow (GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    const gchar * detail,
	    GtkArrowType arrow_type,
	    gint fill, gint x, gint y, gint width, gint height)
{
    cairo_t *cr;

    debug_msg ("draw_arrow",detail,x,y,width,height);

    cr=cairo_create ();			
    cge_gdk (cr,window);

    cge_draw_arrow_from_rc (cr,
			    style,
			    state_type,
			    detail,
			    style->text[state_type],
			    1,
			    x,y,width,height,
			    TRUE,1.5,arrow_type);
  
    cairo_destroy (cr);
}

static void
draw_focus (GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    const gchar * detail, gint x, gint y, gint width, gint height)
{
    cairo_t *cr;
    cge_style *cstyle;

    debug_msg ("draw_focus",detail,x,y,width,height);

    cstyle=cge_get_style (style,detail);

    cr=cairo_create ();		
	
    cge_gdk (cr,window);

    if (!GTK_IS_NOTEBOOK (widget))
	cge_draw_box (cr,
		      style->dark[state_type],
		      NULL,
		      cstyle->line_style.alpha,
		      x+2,y+2,width-4,height-4,
		      cstyle->line_style.round_width,
		      FALSE,
		      cstyle->line_style.line_width/2);
 
    cairo_destroy (cr);
}

static void
draw_box (GtkStyle * style,
	  GdkWindow * window,
	  GtkStateType state_type,
	  GtkShadowType shadow_type,
	  GdkRectangle * area,
	  GtkWidget * widget,
	  const gchar * detail, gint x, gint y, gint width, gint height)
{ 
    cairo_t *cr;
    gint x_off=0,y_off=0;

    debug_msg ("draw_box",detail,x,y,width,height);

    cr=cairo_create ();
  
    cge_gdk (cr,window);
    cairo_save (cr);
    gdk_get_off (window,&x_off,&y_off);
    /* cairo_translate (cr,- (double)x_off,- (double)y_off); */
 
    if (DETAIL ("trough"))
    {
	cge_draw_rectangle (cr,style->bg[GTK_STATE_NORMAL],1,x,y,width,height); 
    }

    if (DETAIL ("menu")
	) {
	gint d;
	/* Don't ask me about that ... */
	gdk_window_get_geometry         (window,
					 &x,
					 &y,
					 &width,
					 &height,
					 &d);
	debug_msg ("draw_box[recalculated]",detail,x,y,width,height);
	/*       cge_draw_box (cr,
		 style->dark[state_type],
		 NULL,
		 1,
		 x,y,width,height,
		 0,
		 FALSE,
		 1);
	*/

	cge_draw_unfilled_box_from_rc (cr,
				       style,
				       state_type,
				       detail,
				       x,
				       y,
				       width,
				       height);
    }
    else 
    {
	cge_draw_filled_box_from_rc (cr,
				     style,
				     state_type,
				     detail,
				     x_off,
				     y_off,
				     x,
				     y,
				     width,
				     height);
      
	cge_draw_unfilled_box_from_rc (cr,
				       style,
				       state_type,
				       detail,
				       x,
				       y,
				       width,
				       height);
      
    }
    cairo_restore (cr);     
    cairo_destroy (cr);
}

static void
draw_box_gap (GtkStyle *style,
	      GdkWindow *window,
	      GtkStateType state_type,
	      GtkShadowType shadow_type,
	      GdkRectangle *area,
	      GtkWidget *widget,
	      const gchar *detail,
	      gint x,
	      gint y,
	      gint width,
	      gint height,
	      GtkPositionType gap_side,
	      gint gap_x,
	      gint gap_width)
{
    cairo_t *cr;
    gint x_off=0,y_off=0;
    GdkColor color;
    cairo_pattern_t *gradient;
    cge_style *cstyle;

    debug_msg ("draw_box_gap",detail,x,y,width,height);

    cstyle=cge_get_style (style,detail);

    cr=cairo_create ();			

    cge_gdk (cr,window);

    gdk_get_off (window,&x_off,&y_off);

    gradient=cge_create_gradient_from_rc (cr,
					  style,
					  state_type,
					  detail,
					  x_off,y_off,width,height); 
    /*
      printf ("detail : %s\n",detail);
      cge_draw_rectangle (cr,style->bg[state_type],1,x,y,width,height); 
    */

    if (cstyle->line_style.color_default[state_type]==TRUE)
	color=style->dark[state_type];
    else
	color=cstyle->line_style.color[state_type];

    cge_draw_box_gap (cr,
		      style->bg[state_type],
		      NULL,
		      cstyle->fill_style.alpha,
		      x,y,width,height,
		      cstyle->fill_style.round_width,
		      gap_x,gap_width,
		      TRUE,0);

    cge_draw_box_gap (cr,
		      color,
		      NULL,
		      cstyle->line_style.alpha,
		      x,y,width,height,
		      cstyle->line_style.round_width,
		      gap_x,gap_width,
		      FALSE,
		      cstyle->line_style.line_width);
  
    if (gradient)
	cairo_pattern_destroy (gradient);
  
    cairo_destroy (cr);
}

#if 0
static void
draw_flat_box (GtkStyle * style,
	       GdkWindow * window,
	       GtkStateType state_type,
	       GtkShadowType shadow_type,
	       GdkRectangle * area,
	       GtkWidget * widget,
	       const gchar * detail, gint x, gint y, gint width, gint height)
{
    cairo_t *cr;
    gint x_off=0,y_off=0;

    debug_msg ("draw_flat_box",detail,x,y,width,height);

    cr=cairo_create ();			
    cge_gdk (cr,window);

    gdk_get_off (window,&x_off,&y_off);
  
    cge_draw_rectangle (cr,style->bg[state_type],1,x,y,width,height);

    cge_draw_filled_box_from_rc (cr,
				 style,
				 state_type,
				 detail,
				 x_off,
				 y_off,
				 x,
				 y,
				 width,
				 height);

    cge_draw_unfilled_box_from_rc (cr,
				   style,
				   state_type,
				   detail,
				   x,
				   y,
				   width,
				   height);
    

    cairo_destroy (cr);
}
#endif

static void
draw_check (GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    const gchar * detail, 
	    gint x, gint y, gint width, gint height)
{
    cairo_t *cr;
    cge_style *cstyle;

    debug_msg ("draw_check",detail,x,y,width,height);

    cstyle=cge_get_style (style,detail);

    cr=cairo_create ();			
    cge_gdk (cr,window);

    if (DETAIL ("check")) {
	cge_draw_box (cr,
		      style->bg[state_type],
		      NULL,
		      1,
		      x,y,width,height,
		      cstyle->fill_style.round_width,
		      TRUE,0);    
	if (shadow_type == GTK_SHADOW_IN) {
	    /* here vectored check */
	    cge_draw_checked (cr,
			      style->text[state_type],			   
			      1,
			      x+2,y+2,width-4,height-4,
			      cstyle->line_style.line_width);
	}
    }
    else
    {
	cge_draw_box (cr,
		      style->base[state_type],
		      NULL,
		      1,
		      x,y,width,height,
		      cstyle->fill_style.round_width,
		      TRUE,0);    
      
	cge_draw_unfilled_box_from_rc (cr,
				       style,
				       state_type,
				       detail,
				       x,
				       y,
				       width,
				       height);
  
	if (shadow_type == GTK_SHADOW_IN) {
	    /* here vectored check */
	    cge_draw_checked (cr,
			      style->text[state_type],			   
			      1,
			      x+2,y+2,width-4,height-4,
			      cstyle->line_style.line_width);
		
	}
    }

    cairo_destroy (cr);
}

/* FIXME : make a true circle draw function */
static void
draw_option (GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     const gchar * detail, gint x, gint y, gint width, gint height)
{
    cairo_t *cr;
    cge_style *cstyle;

    debug_msg ("draw_option",detail,x,y,width,height);

    cstyle=cge_get_style (style,detail);

    cr=cairo_create ();			
    cge_gdk (cr,window);
  
    /* cge_draw_rectangle (cr,style->bg[state_type],1,x,y,width,height); */

    if (DETAIL ("option")) {
	cge_draw_circle (cr,
			 style->bg[state_type],
			 NULL,
			 1,
			 x,y,
			 width/2,
			 TRUE,0);    
	if (shadow_type == GTK_SHADOW_IN) {
	    /* here vectored option */
	    cge_draw_circle (cr,
			     style->fg[state_type],
			     NULL,
			     1,
			     x+1,y+1,
			     width/2-2,
			     TRUE,0);    
	}
    }
    else
    {
	cge_draw_circle (cr,
			 style->base[state_type],
			 NULL,
			 1,
			 x,y,
			 width/2,
			 TRUE,0);    
	cge_draw_circle (cr,
			 style->dark[state_type],
			 NULL,
			 1,
			 x,y,
			 width/2,
			 FALSE,
			 cstyle->line_style.line_width);

  
	if (shadow_type == GTK_SHADOW_IN) {
	    /* here vectored option */
	    cge_draw_circle (cr,
			     style->text[state_type],
			     NULL,
			     1,
			     x+2,y+2,
			     (width)/2 - 2,
			     TRUE,0);    
	}
    }

    cairo_destroy (cr);
}

/* tab */
static void
draw_extension (GtkStyle * style,
		GdkWindow * window,
		GtkStateType state_type,
		GtkShadowType shadow_type,
		GdkRectangle * area,
		GtkWidget * widget,
		const gchar * detail,
		gint x, gint y, gint width, gint height,
		GtkPositionType gap_side)
{
    cairo_t *cr;
    cairo_pattern_t *gradient;
    GdkColor color;
    gint x_off,y_off;
    cge_style *cstyle;

    debug_msg ("draw_extension",detail,x,y,width,height);

    cstyle=cge_get_style (style,detail);

    cr=cairo_create ();			
    cge_gdk (cr,window);

    gdk_get_off (window,&x_off,&y_off);

    gradient=cge_create_gradient_from_rc (cr,
					  style,
					  state_type,
					  detail,
					  x_off,y_off,width,height); 

    /* cge_draw_rectangle (cr,style->bg[GTK_STATE_NORMAL],1,x,y,width,height); */
	
    if (cstyle->line_style.color_default[state_type]==TRUE)
	color=style->dark[state_type];
    else
	color=cstyle->line_style.color[state_type];

    cge_draw_notebook_tab (cr,
			   style->bg[state_type],
			   gradient,
			   cstyle->fill_style.alpha,
			   x,y,width,height,
			   cstyle->fill_style.round_width,
			   TRUE,0);
    cge_draw_notebook_tab (cr,
			   color,
			   NULL,
			   cstyle->line_style.alpha,
			   x,y,width,height,
			   cstyle->line_style.round_width,
			   FALSE,
			   cstyle->line_style.line_width);

    if (gradient)
	cairo_pattern_destroy (gradient);
  
    cairo_destroy (cr);

}

static void
draw_slider (GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     const gchar * detail,
	     gint x, gint y, gint width, gint height,
	     GtkOrientation orientation)
{
    cairo_t *cr;
    gint x_off=0,y_off=0;

    debug_msg ("draw_slider",detail,x,y,width,height);

    cr=cairo_create ();			
    cge_gdk (cr,window);

    gdk_get_off (window,&x_off,&y_off);

    cge_draw_filled_box_from_rc (cr,
				 style,
				 state_type,
				 detail,
				 x_off,
				 y_off,
				 x,
				 y,
				 width,
				 height);

    cge_draw_unfilled_box_from_rc (cr,
				   style,
				   state_type,
				   detail,
				   x,
				   y,
				   width,
				   height);
  

    /*
      cge_draw_slider (cr,
      style->dark[state_type],
      1,
      x+4,y+height/2 - height/8,width-8,height/4,
      THEME_DATA (style)->line_style.line_width,3);
    */
    cairo_destroy (cr);
}

static void
draw_handle (GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     const gchar * detail,
	     gint x, gint y, gint width, gint height,
	     GtkOrientation orientation)
{
    cairo_t *cr;
    gint x_off=0,y_off=0;

    debug_msg ("draw_handle",detail,x,y,width,height);

    cr=cairo_create ();			
    cge_gdk (cr,window);

    gdk_get_off (window,&x_off,&y_off);

    /* cge_draw_rectangle (cr,style->bg[GTK_STATE_NORMAL],1,x,y,width,height); */

    cge_draw_filled_box_from_rc (cr,
				 style,
				 state_type,
				 detail,
				 x_off,
				 y_off,
				 x,
				 y,
				 width,
				 height);

    cge_draw_unfilled_box_from_rc (cr,
				   style,
				   state_type,
				   detail,
				   x,
				   y,
				   width,
				   height);

  
    cairo_destroy (cr);

}

static void
draw_resize_grip (GtkStyle * style, 
		  GdkWindow * window,
		  GtkStateType state_type, 
		  GdkRectangle * area,
		  GtkWidget * widget, const gchar * detail,
		  GdkWindowEdge edge, gint x, gint y, gint width, gint height)
{
    cairo_t *cr;
    gint x_off=0,y_off=0;

    debug_msg ("draw_resize_grip",detail,x,y,width,height);

    cr=cairo_create ();			
    cge_gdk (cr,window);

    gdk_get_off (window,&x_off,&y_off);

    /* cge_draw_rectangle (cr,style->bg[GTK_STATE_NORMAL],1,x,y,width,height); */

    cge_draw_filled_box_from_rc (cr,
				 style,
				 state_type,
				 detail,
				 x_off,
				 y_off,
				 x+1,
				 y+1,
				 width-2,
				 height-2);

    cge_draw_unfilled_box_from_rc (cr,
				   style,
				   state_type,
				   detail,
				   x+1,
				   y+1,
				   width-2,
				   height-2);

    cairo_destroy (cr);

}

/* should work */
#if 0
static void
set_background (GtkStyle      *style,
		GdkWindow     *window,
		GtkStateType   state_type)
{
    gint x,y,w,h,d;
    gint x_off=0,y_off=0;
    cairo_t *cr;
    GdkPixmap *pm;
    GdkRectangle rect;

    gdk_window_get_geometry         (window,
				     &x,
				     &y,
				     &w,
				     &h,
				     &d);
    rect.x=x;
    rect.y=y;
    rect.width=w;
    rect.height=h;
 
    cr=cairo_create ();			

    pm = gdk_pixmap_new (window,w,h,d);
    cge_gdk (cr,pm);

    gdk_window_begin_paint_rect     (window,
				     &rect);
    gdk_get_off (window,&x_off,&y_off);
    gdk_window_end_paint            (window);

    cairo_translate (cr,- (double)x_off,- (double)y_off);


    cge_draw_rectangle (cr,style->bg[GTK_STATE_NORMAL],1,x,y,w,h);
    /* we can now have style defined background ! */
    /*
      cge_draw_filled_box_from_rc (cr,
      style,
      state_type,
      "background",
      x_off,
      y_off,
      x,
      y,
      w,
      h);
  
    */
    gdk_window_set_back_pixmap (window,pm,FALSE);  

    /* g_free (pm); */

    cairo_destroy (cr);
}
#endif

GType cge_type_style = 0;

void
cge_style_register_type (GTypeModule * module)
{
    static const GTypeInfo object_info = {
	sizeof (CgeStyleClass),
	(GBaseInitFunc) NULL,
	(GBaseFinalizeFunc) NULL,
	(GClassInitFunc) cge_style_class_init,
	NULL,		/* class_finalize */
	NULL,		/* class_data */
	sizeof (CgeStyle),
	0,		/* n_preallocs */
	(GInstanceInitFunc) cge_style_init,
    };
  
    cge_type_style = g_type_module_register_type (module,
						  GTK_TYPE_STYLE,
						  "CgeStyle",
						  &object_info, 0);
}

static void
cge_style_init (CgeStyle * style)
{

}

static void
cge_style_class_init (CgeStyleClass * klass)
{
    GtkStyleClass *style_class = GTK_STYLE_CLASS (klass);
  
    parent_class = g_type_class_peek_parent (klass);

    style_class->draw_hline = draw_hline;
    style_class->draw_vline = draw_vline;  
    style_class->draw_shadow = draw_shadow; 
    style_class->draw_arrow = draw_arrow;
    style_class->draw_tab = draw_tab;
    style_class->draw_focus = draw_focus;  
    style_class->draw_box = draw_box;
    /*  style_class->draw_flat_box = draw_flat_box; */
    style_class->draw_box_gap = draw_box_gap;
    style_class->draw_extension = draw_extension;
    style_class->draw_slider = draw_slider;  
    style_class->draw_handle = draw_handle;  
    style_class->draw_resize_grip = draw_resize_grip;  
    style_class->draw_shadow_gap = draw_shadow_gap;  
  
    style_class->draw_check = draw_check;
    style_class->draw_option = draw_option;  

    /*
      style_class->draw_polygon = draw_polygon;
      style_class->draw_diamond = draw_diamond;  
      style_class->draw_layout = draw_layout;   
      style_class->set_background=set_background;
    */
}
