#ifndef _CGE_STYLE_H_
#define _CGE_STYLE_H_

typedef struct _cge_style cge_style;
typedef struct _cge_fill_style cge_fill_style;
typedef struct _cge_line_style cge_line_style;

typedef struct _CgeStyle CgeStyle;
typedef struct _CgeStyleClass CgeStyleClass;
typedef struct _CgeRCStyle CgeRCStyle;
typedef struct _CgeRCStyleClass CgeRCStyleClass;

extern GType cge_type_style;
extern GType cge_type_rc_style;

#define CGE_TYPE_STYLE              cge_type_style
#define CGE_STYLE(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), CGE_TYPE_STYLE, CgeStyle))
#define CGE_STYLE_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), CGE_TYPE_STYLE, CgeStyleClass))
#define CGE_IS_STYLE(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), CGE_TYPE_STYLE))
#define CGE_IS_STYLE_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), CGE_TYPE_STYLE))
#define CGE_STYLE_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), CGE_TYPE_STYLE, CgeStyleClass))
#define CGE_TYPE_RC_STYLE              cge_type_rc_style
#define CGE_RC_STYLE(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), CGE_TYPE_RC_STYLE, CgeRCStyle))
#define CGE_RC_STYLE_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), CGE_TYPE_RC_STYLE, CgeRCStyleClass))
#define CGE_IS_RC_STYLE(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), CGE_TYPE_RC_STYLE))
#define CGE_IS_RC_STYLE_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), CGE_TYPE_RC_STYLE))
#define CGE_RC_STYLE_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), CGE_TYPE_RC_STYLE, CgeRCStyleClass))

#define CGE_RC_DATA(rc_style)          (CGE_RC_STYLE (rc_style))

#define THEME_DATA(style)                 (CGE_RC_STYLE (style->rc_style))
#define NEW_THEME_DATA(rc_style)          (CGE_RC_STYLE (rc_style))
#define SET_THEME_DATA(rc_style, data)    (CGE_RC_DATA (rc_style))


typedef gboolean (*CgeTranslateEnumFunc) (gchar * str, gint *retval);

#define CGE_TYPE_COUNT 6

typedef enum 
  {
    CGE_FILL_SOLID=0,
    CGE_FILL_GRADIENT,    
    CGE_FILL_FILE    
  }cge_fill_mode;

typedef enum 
  {
    CGE_LINE_SOLID=0,
    CGE_LINE_NONE,
    CGE_LINE_DASHED,
    CGE_LINE_BEVEL,
    CGE_LINE_FILE
  }cge_line_mode;


typedef enum
  {
    CGE_DIRECTION_VERTICAL=0,
    CGE_DIRECTION_HORIZONTAL    
  }cge_direction;

typedef enum
  {
    CGE_ARROW_NORMAL=0,
    CGE_ARROW_SIMPLE,
    CGE_ARROW_WONDERLAND,
    /* others ... */
    CGE_ARROW_FILE

  }cge_arrow_style;

typedef enum
  {
    CGE_CHECK_NORMAL=0,
    CGE_CHECK_SIMPLE,
    CGE_CHECK_WONDERLAND,
    /* others ... */
    CGE_CHECK_FILE

  }cge_check_style;

typedef enum
  {
    CGE_OPTION_NORMAL=0,
    CGE_OPTION_SIMPLE,
    CGE_OPTION_WONDERLAND,
    /* others ... */
    CGE_OPTION_FILE

  }cge_option_style;

#define MAX_GRADIENT_COLORS 10
#define MAX_STYLES 10



struct _cge_fill_style
{
  cge_fill_mode fill_mode;
  
  cge_direction gradient_direction;
  gint gradient_colors_count[5];
  gboolean gradient_colors_default[5];
  GdkColor gradient_colors[5][MAX_GRADIENT_COLORS];
  gdouble round_width;
  gdouble alpha;
};

struct _cge_line_style
{
  cge_line_mode line_mode;

  gdouble line_width;
  gdouble round_width;

  gboolean color_default[5];
  GdkColor color[5];
  gdouble alpha;
};

struct _cge_style
{
  gchar *detail;

  cge_fill_style fill_style;
  cge_line_style line_style;
};


struct _CgeStyle {
  GtkStyle parent_instance;
  

};

struct _CgeStyleClass {
  GtkStyleClass parent_class;
};

struct _CgeRCStyle {
  GtkRcStyle parent_instance;



  cge_arrow_style arrow_style;

  cge_style styles[MAX_STYLES];

};

struct _CgeRCStyleClass {
  GtkRcStyleClass parent_class;
};

void cge_style_register_type (GTypeModule * module);
void cge_rc_style_register_type (GTypeModule * module);

#endif
